//
// SubSurface class implementation
//

#include <iostream>
#include <fstream>
#include <sstream>
#include <queue>

#include "subsurface.h"

using namespace std;

float avgR = 0;
float underLu = 0;
float under10Lu = 0;
float cnt = 0;

/** constructor */
SubSurface::SubSurface(): eta(0), F_dr(0), F_dt(0), A(0), summarized(false),
    sampleFactor(1)
{
    for (int color=0; color < 3; color++) {
        samples[color] = SSOctree(8);
    }
}


/** calculate radiant exitance */
Spectrum SubSurface::Lo(const Point &p, const Normal &n, const Vector &wo) {
//    cout << endl;
//    samples[0].print();
//    exit(0);
    
    // traverse down octree to find Mo, for each color
//    avgR = 0;
//    cnt = 0;
//    underLu = 0;
//    under10Lu = 0;
    
    Spectrum Mo;
    vector<ESample> samps;
    for (int color = 0; color < 3; color++) {
        if (!samples[color].empty()) {
//          Mo.c[color] = traverseMo(&samples[color].root, p, n, wo, samps).c[color];
            Mo = traverseMo(&samples[color].root, p, n, wo, samps);
        }
    }

//    cout << endl;
//    cout << "P: " << p << endl;
//    cout << "Samps: " << samps.size() << endl;
//    for (size_t i=0; i < samps.size(); i++) {
//        cout << "r: " << (p - samps[i].p).Length() << " " << samps[i] << endl;
//    }
//    exit(0);
    
//    cout << "Avg r: " << avgR / cnt << " cnt: " << cnt << " underLu: " << underLu;
//    cout << " under10Lu: " << under10Lu << endl;
    
    //TODO use Ft/Fdr?
    Spectrum Lo = Mo / M_PI;

    return Lo;
}

/** traverse tree to find radiance */
Spectrum SubSurface::traverseMo(SSOctNode* node, const Point &p, 
                                const Normal &n, const Vector &wo,
                                vector<ESample>& samps) 
{
    Spectrum Mo;
    
    // octree calc
    float dW = node->deltaW(p);
    
    //TODO if inside node, eval children?
    //TODO EPSILON value?
    float EPSILON  = 0.05f;

    if (!node->data.empty()) {
        // node is a leaf, just use the samples
        for (size_t i=0; i < node->data.size(); i++) {
            Mo += node->data[i].Mo(p, n, wo);
        }
    } else if (dW < EPSILON) {
        // use node directly
        Mo += node->summary.Mo(p, n, wo);
    } else {
        // use children
        for (int child = 0; child < 8; ++child) {
            if (node->children[child]) {
                Mo += traverseMo(node->children[child], p, n, wo, samps);
            }
        }
    }

    // find all samples within dist of p
//    float dist = lu.c[0] * 20.f;
//    BBox maxbox(p - Vector(dist,dist,dist), p + Vector(dist, dist, dist));
//    if (node->bound.Overlaps(maxbox)) {
//        if (maxbox.Inside(node->bound.pMin) && maxbox.Inside(node->bound.pMax)) {
//            // node entirely inside box, get samples if leaf else go on
//            if (!node->data.empty()) {
//                for (size_t i=0; i < node->data.size(); i++) {
//                    //samps.push_back(node->data[i]);
//                    Mo += node->data[i].Mo(p, n, wo);
//                }
//            } else {
//                for (int child = 0; child < 8; ++child) {
//                    if (node->children[child]) {
//                        Mo += traverseMo(node->children[child], p, n, wo, samps);
//                    }
//                }
//            }
//        } else {
//            // not entirely inside, recurse to children that overlap
//            for (int child = 0; child < 8; ++child) {
//                if (node->children[child] && 
//                    node->children[child]->bound.Overlaps(maxbox)) {
//                    Mo += traverseMo(node->children[child], p, n, wo, samps);
//                }
//            }
//        }
//    }


    return Mo;
}


/** calculate radiant exitance */
Spectrum ESample::Mo(const Point &xo, const Normal &n, const Vector &wo) {
    Spectrum Mo;
    
    float r = (xo - p).Length();
    
    
    // ind calc
//    for (int color = 0; color < 3; color++) {
//        float sig_tr = sqrt( 3 * subSurf->sigma_a.c[color] * subSurf->sigma_tp.c[color]);
//        float zr = subSurf->lu.c[color];
//        float zv = zr * (1 + (4.f / 3.f) * subSurf->A);
//        float dr = sqrt( (r * r) + (zr * zr) );
//        float dv = sqrt( (r * r) + (zv * zv) );
//        
//        float c1 = zr * (sig_tr + (1.f / dr));
//        float c2 = zv * (sig_tr + (1.f / dv));
//        float dmo = (c1 * exp(-sig_tr * dr) / (dr * dr)) +
//                    (c2 * exp(-sig_tr * dv) / (dv * dv));
//        dmo = dmo / (4.f * M_PI);
//        
//        Mo.c[color] = subSurf->F_dt * dmo * Ep.c[color] * Ap;
//    }

    
    //spectrum calc
//    int num = 60;
//    for (int i=0; i < num; i++) {
//        r = (float)i * 0.2;
        
//    if (r < subSurf->lu.c[2]) {
//        underLu += 1;
//    }
//    if (r < 10.f * subSurf->lu.c[2]) {
//        under10Lu += 1;
//    }
//
//    avgR += r;
//    cnt += 1;
    
    Spectrum z_r2 = subSurf->z_r * subSurf->z_r;
    Spectrum dr = (Spectrum(r * r) + z_r2).Sqrt();
    Spectrum z_v2 = subSurf->z_v * subSurf->z_v;
    Spectrum dv = (Spectrum(r * r) + z_v2).Sqrt();
    
    Spectrum C1 = subSurf->z_r * (subSurf->sigma_tr + (Spectrum(1.0) / dr)); 
    Spectrum C2 = subSurf->z_v * (subSurf->sigma_tr + (Spectrum(1.0) / dv));
    
    Spectrum dMo1 = C1 * Exp(-1 * subSurf->sigma_tr * dr) / (dr * dr);
    Spectrum dMo2 = C2 * Exp(-1 * subSurf->sigma_tr * dv) / (dv * dv);
    
    Spectrum dMo = (1.f / (4.f * M_PI)) * (dMo1 + dMo2);
    
    Mo = subSurf->F_dt * dMo * Ep * Ap;
        
//        cout << "r: " << r << " dMo: " << dMo  << " z_r2: " << z_r2;
//        cout << " dr: " << dr<< " z_v2: " << z_v2<< " dv: " << dv ;
//        cout << " C1: " << C1<<" C2: " << C2<<" dr: " << dr << endl;
//    }
//    exit(0);

    // TODO: don't eval points 1/sigma_tp or closer?
//    for (int col=0; col < 3; col++) {
//        if (r < subSurf->lu.c[col]) {
//            Mo.c[col] = 0;
//        }
//    }
    
    return Mo;
    
}

/** set scattering params */
void SubSurface::setParams(vector<float>& vals) {
    // store params
    float tmp[3];
    tmp[0] = vals[0];
    tmp[1] = vals[1];
    tmp[2] = vals[2];
    sigma_sp = Spectrum(tmp);
    tmp[0] = vals[3];
    tmp[1] = vals[4];
    tmp[2] = vals[5];
    sigma_a = Spectrum(tmp);
    eta = vals[6];
    if (vals.size() > 6) {
        sampleFactor = vals[7];
    }
    
    // calc constants
    F_dr = (-1.440 / (eta * eta)) + (0.710 / eta) + 0.668 + (0.0636 * eta);
    F_dt = 1.f - F_dr;
    A = (1.0 + F_dr) / (1.0 - F_dr);
    sigma_tp = sigma_sp + sigma_a;
    lu = Spectrum(1.f) / sigma_tp;
    
    sigma_tr = (3 * sigma_a * sigma_tp).Sqrt();
    z_r = Spectrum(1.f) / sigma_tp;
    z_v = z_r * (1 + (4.0/3.0) * A);
    
    Spectrum ap = (sigma_sp / sigma_tp);
    Spectrum Rd = (ap / 2.f) * (Spectrum(1.f) + Exp((-3.f / 4.f) * A * (3.f * (Spectrum(1.f) - ap)).Sqrt())) * Exp(-(3.f * (Spectrum(1.f) - ap)).Sqrt());
    
    cout << " sigma_sp: " << sigma_sp << " sigma_a: " << sigma_a;
    cout << " sigma_tp: " << sigma_tp;
    cout << " eta: " << eta << " sampleFactor: " << sampleFactor << endl;
    cout << " F_dr: " << F_dr << endl;
    cout << " sigma_tr: " << sigma_tr << endl;
    cout << " A: " << A << endl;
    cout << " lu: " << lu << endl;
    cout << " ap: " << ap << endl;
    cout << " Rd: " << Rd << endl;
}

/** add an irradiance sample */
void SubSurface::addSample(Spectrum& e, float area, const Point& loc, int color) {
    // check for NAN
    for (int col=0; col < 3; col++) {
        if (isnan(e.c[col])) {
            cout << "NAN in Sample! p: " << loc << " e: " << e << " area: " << area  << endl;
            return;
        }
    }
    ESample samp(e, area, loc, this);
    samples[color].Add(samp, samp.getBBox());
}
    
/** add a primitive */
void SubSurface::addPrim(Reference<Primitive> p) {
    prims.push_back(p);
    for (int color = 0; color < 3; color++) {
        samples[color].bound = Union(samples[color].bound, p->WorldBound());
    }
}

/** get bounding box for sample */
BBox ESample::getBBox() {
    // TODO: bound with bigger box?
    BBox box(p);
    return box;
}

/** summarize samples */
void SubSurface::summarize() {
    if (!summarized) {
        summarized = true;
        for (int color = 0; color < 3; color++) {
            if (!samples[color].empty()) {
                samples[color].root.summarize(this, color);
                //samples[color].print();
            }
        }
    }
    
}

/* whether tree is empty */
bool SSOctree::empty() {
    for (int i=0; i < 8; i++) {
        if (root.children[i]) {
            return false;
        }
    }
    return true;
}
    
    
/** get solid angle spanned by points */
float SSOctNode::deltaW(const Point& x) {
    //cout << "x = " << x << " sum = " << summary << endl;
    return summary.Ap / (x - summary.p).LengthSquared();
}

/** position weight */
float ESample::getPositionWeight(int color) {
    //return Ep.c[2] * Ap;
    return (Ep.c[0] + Ep.c[1] + Ep.c[2]) * Ap;
}
    

/** summarize children */
void SSOctNode::summarize(SubSurface* ss, int color) {
    //TODO using y to scale weight correct?
    
    summary.subSurf = ss; // ptr to ss obj
    float totalWeight = 0;
    float numPts = 0;
    Point totalP(0,0,0); 
    if (data.empty()) {
        // sum of children summaries
        for (int child = 0; child < 8; ++child) {
            if (children[child]) {
                children[child]->summarize(ss, color);
                summary.Ep += children[child]->summary.Ep * children[child]->summary.Ap;
                summary.Ap += children[child]->summary.Ap;
                float weight = children[child]->summary.getPositionWeight(color);
                summary.p += children[child]->summary.p * weight;
                totalWeight += weight;
                numPts++;
                totalP += children[child]->summary.p;
            }
        }
    }
    else {
        // sum of data
        for (size_t i = 0; i < data.size(); ++i) {
            summary.Ep += data[i].Ep * data[i].Ap;
            summary.Ap += data[i].Ap;
            float weight = data[i].getPositionWeight(color);
            summary.p += data[i].p * weight;
            totalWeight += weight;
            numPts++;
            totalP += data[i].p;
        }
    }
    // store the average irradiance?
    summary.Ep /= summary.Ap;
    
    // average the position, careful of zero weights
    if (isnan(summary.p.x)) {
        cout << "NAN in Summary Before Divide! " << summary << endl;
        exit(0);
    }
    Point beforeDivide = summary.p;
    if (totalWeight == 0.f) {
        // just average all positions equally in this case
        summary.p = totalP / numPts;
    } else {
        summary.p /= totalWeight;
    }
    if (isnan(summary.p.x)) {
        cout << "NAN in Summary! " << summary << endl;
        if (numPts > 0) {
            summary.p = beforeDivide / numPts;
        } else {
            cout << "numPts Zero! " << summary << endl;
        } 
    }
}

/** add sample */
void SSOctree::addPrivate(SSOctNode *node, const BBox nodeBound,
                        ESample dataItem, const BBox dataBound,
                        float diag2, int depth) {
    // set bounds for current node
    node->bound = nodeBound;
                                
    // Possibly add data item to current octree node
    if (depth == maxDepth ||
        DistanceSquared(nodeBound.pMin,
                        nodeBound.pMax) < diag2) {
        node->data.push_back(dataItem);
        return;
    }
    // Otherwise add data item to octree children
    Point pMid = .5 * nodeBound.pMin + .5 * nodeBound.pMax;
    // Determine which children the item overlaps
    bool over[8];
    over[0] = over[1] =
              over[2] =
              over[3] = (dataBound.pMin.x <= pMid.x);
    over[4] = over[5] =
              over[6] =
              over[7] = (dataBound.pMax.x  > pMid.x);
    over[0] &= (dataBound.pMin.y <= pMid.y);
    over[1] &= (dataBound.pMin.y <= pMid.y);
    over[4] &= (dataBound.pMin.y <= pMid.y);
    over[5] &= (dataBound.pMin.y <= pMid.y);
    over[2] &= (dataBound.pMax.y  > pMid.y);
    over[3] &= (dataBound.pMax.y  > pMid.y);
    over[6] &= (dataBound.pMax.y  > pMid.y);
    over[7] &= (dataBound.pMax.y  > pMid.y);
    over[0] &= (dataBound.pMin.z <= pMid.z);
    over[2] &= (dataBound.pMin.z <= pMid.z);
    over[4] &= (dataBound.pMin.z <= pMid.z);
    over[6] &= (dataBound.pMin.z <= pMid.z);
    over[1] &= (dataBound.pMax.z  > pMid.z);
    over[3] &= (dataBound.pMax.z  > pMid.z);
    over[5] &= (dataBound.pMax.z  > pMid.z);
    over[7] &= (dataBound.pMax.z  > pMid.z);
    for (int child = 0; child < 8; ++child) {
        if (!over[child]) continue;
        if (!node->children[child]) {
            node->children[child] = new SSOctNode;
            node->children[child]->depth = depth + 1;
        }
        // Compute _childBound_ for octree child _child_
        BBox childBound;
        childBound.pMin.x = (child & 4) ? pMid.x : nodeBound.pMin.x;
        childBound.pMax.x = (child & 4) ? nodeBound.pMax.x : pMid.x;
        childBound.pMin.y = (child & 2) ? pMid.y : nodeBound.pMin.y;
        childBound.pMax.y = (child & 2) ? nodeBound.pMax.y : pMid.y;
        childBound.pMin.z = (child & 1) ? pMid.z : nodeBound.pMin.z;
        childBound.pMax.z = (child & 1) ? nodeBound.pMax.z : pMid.z;
        addPrivate(node->children[child], childBound,
                   dataItem, dataBound, diag2, depth+1);
    }
}

/* debug */
void SSOctree::print() {
    
    //print(&root, 0);
    
    queue<SSOctNode*> vec;
    vec.push(&root);
    while (!vec.empty()) {
        SSOctNode* node = vec.front();
        vec.pop();
        for (int i=0; i < node->depth; i++) {
            cout << " ";
        }
        cout << "Bound: " << node->bound << " ";
        if (node->data.empty()) {
            cout << "Sum: " << node->summary << endl;
        } else {
            for (size_t i=0; i < node->data.size(); i++) {
                cout << "Data: " << node->data[i] << endl;
            }
        }
        for (int i=0; i < 8; i++) {
            if (node->children[i]) {
                vec.push(node->children[i]);
            }
        }
    }
}

/* debug */
void SSOctree::print(SSOctNode* node, int depth) {
    for (int s=0; s < depth; s++) {
        cout << " ";
    }
    
    if (node->data.empty()) {
        cout << "Sum: " << node->summary << endl;
    } else {
        for (size_t i=0; i < node->data.size(); i++) {
            cout << "Data: " << node->data[i] << endl;
        }
    }
    for (int i=0; i < 8; i++) {
        if (node->children[i]) {
            print(node->children[i], depth + 1);
        }
    }
}

/** get the global map of subsurface objects */
map<int, SubSurface>& getSubSurfaceMap() {
    static map<int, SubSurface> subSurfMap;
    return subSurfMap;
}

/** method to add a primitive to a subsurface obj according to id */
void addSubSurfacePrimitive(Reference<Primitive> p) {
    static map<int, string> names;
    static map<int, vector<float> > params;
    // read params if first time 
    if (params.empty()) {
        ifstream file("ss.conf");
        if (file.fail()) {
            Error("Subsurface param file ss.conf not found! ");
            exit(0);
        }
        string line;
        stringstream sst;
        while (getline(file, line)) {
            if (!line.empty() && line[0] != '#') {
                sst.clear();
                sst.str(line);
                int sid;
                string name;
                float val;
                sst >> sid;
                sst >> name;
                names[sid] = name;
                while (sst >> val) {
                    params[sid].push_back(val);
                }                
            }
        }
    }     
    // set params on subsurf obj if new
    int id = p->subSurfaceId;
    if (getSubSurfaceMap().find(id) == getSubSurfaceMap().end()) {
        cout << names[id] << " ssid: " << id << endl;
        getSubSurfaceMap()[id].setParams(params[id]);
    }
    // fully refine primitive
    vector< Reference<Primitive> > prims;
    p->FullyRefine(prims);
    // add primitives and update area
    for (size_t i = 0; i < prims.size(); i++) {
        Reference<Primitive> *rp = &prims[i];
        Reference<GeometricPrimitive> *rgp = (Reference<GeometricPrimitive> *) rp;
        Reference<Shape> shape = (*rgp)->getShape();
        getSubSurfaceMap()[id].area += shape->Area();
        getSubSurfaceMap()[id].addPrim(prims[i]);
    }
}
